package com.jozki.uutils.javafx.about;

import java.util.Random;

import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.event.EventHandler;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.text.Font;
import javafx.scene.text.FontPosture;
import javafx.scene.text.FontWeight;
import javafx.stage.Stage;
import javafx.stage.Window;
import javafx.stage.WindowEvent;
import javafx.util.Duration;

public class About extends Stage {

	/**
	 * Configuration class for About window. All values can be null.
	 */
	public static class AboutConfig {
		public String projectTitle;
		public String projectSubtitle;
		public String author;
		/**
		 * The title of some additional text.
		 */
		public String additionalTitle;
		/**
		 * Additional text.
		 */
		public String additionalText;
		public String desc[];
		public String date;
		int numberOfLines;
		/**
		 * Minimal width of line. Can be close to zero.
		 */
		double lineMinWidth;
		double lineMaxWidth;
		/**
		 * How many times per second should lines be updated. Default is 30;
		 */
		int updateFPS;

		void adjustValues() {
			if (numberOfLines <= 0) {
				numberOfLines = 12;
			}
			if (lineMinWidth <= 0) {
				lineMinWidth = 0.5;
			}
			if (lineMaxWidth <= 0) {
				lineMaxWidth = 1;
			}
			if (updateFPS <= 0) {
				updateFPS = 30;
			}
		}
	}

	/**
	 * Creates autopositioned About window.
	 * @param window Parent window to which this window should be positioned.
	 * @param projectTitle Title of the About window.
	 * @param hiddenEvent Event called when About window is being closed (for cleanup purposes).
	 * @param config Configuration for the About window. See {@link AboutConfig}.
	 * @return
	 */
	public static About create(Window window, String projectTitle, EventHandler<WindowEvent> hiddenEvent, AboutConfig config) {

		About stage = new About();
		if (config == null) {
			config = new AboutConfig();
		}
		config.adjustValues();
		stage.config = config;

		Scene scene = new Scene(stage.createContent(), 365, stage.estimateHeight());
		stage.setTitle(projectTitle + " About ..");
		stage.setScene(scene);
		stage.setResizable(false);
		if (window != null) {
			// position in the center of parent
			stage.setX(window.getX() + (window.getWidth() - scene.getWidth()) / 2);
			stage.setY(window.getY() + (window.getHeight() - scene.getHeight()) / 2);
		}
		stage.setOnHidden(hiddenEvent);
		stage.setOnHiding(e -> stage.destroy());
		stage.show();

		return stage;
	}

	private void destroy() {
		timeline.stop();
	}

	/**
	 * Estimates the window height based on visible content 
	 */
	private double estimateHeight() {
		int height = 100;
		if (config.author != null) {
			height += TITLES_GAP;
		}
		if (config.additionalTitle != null) {
			height += TITLES_GAP;
		}
		if (config.desc != null) {
			height += config.desc.length * DESC_GAP;
		}
		return height;
	}

	private AboutConfig config;

	private static final int TITLES_GAP = 55;
	private static final int DESC_GAP = 20;

	private static final Color[] COLORS_POOL = { Color.rgb(170, 255, 0), Color.rgb(146, 220, 0), Color.rgb(122, 185, 0), Color.rgb(98, 150, 0), Color.rgb(74, 115, 0), Color.rgb(50, 80, 0) };

	private static class FunnyLine {
		/**
		 * Position of line.
		 */
		public double x;
		/**
		 * Maximal position of line (when reached position and max is reseted).
		 */
		public double maxX;
		/**
		 * Velocity of funny line.
		 */
		public double vx;
		public Color color = Color.BLACK;
		public double width;

		private AboutConfig config;

		private FunnyLine(AboutConfig config) {
			this.config = config;
		}

		public void updatePosition(double time) {
			x += vx * time;
		}

		public void setRandom(Random r, double parentMaxX) {
			if (parentMaxX <= 0) {
				parentMaxX = 1;
			}
			x = r.nextInt((int) parentMaxX);
			vx = r.nextDouble() * 20 + 5;
			width = r.nextDouble() * (config.lineMaxWidth - config.lineMinWidth) + config.lineMinWidth;
			maxX = r.nextInt((int) parentMaxX);
		}

		public boolean isAlive() {
			return x <= maxX;
		}
	}

	private FunnyLine[] lines;

	private Random r = new Random();

	private Font fontItemName = Font.font("Courier New", 14);
	private Font fontItemValue = Font.font("Arial New", FontWeight.BOLD, 25);
	private Font fontDesc = Font.font("Courier New", FontPosture.ITALIC, 14);
	private Font fontDate = Font.font("Arial New", 16);
	private Font fontSubtitle = Font.font("Arial New", 12);
	private GraphicsContext gc;

	private Canvas canvas;
	private Timeline timeline;

	/**
	 * Creates the content of About window
	 */
	private Parent createContent() {
		canvas = new Canvas();
		Pane group = new Pane();
		group.getChildren().add(canvas);
		canvas.widthProperty().bind(group.widthProperty());
		canvas.heightProperty().bind(group.heightProperty());

		gc = canvas.getGraphicsContext2D();

		lines = new FunnyLine[config.numberOfLines];
		for (int i = 0; i < lines.length; ++i) {
			lines[i] = new FunnyLine(config);
			lines[i].setRandom(r, canvas.getWidth());
			lines[i].color = COLORS_POOL[i % COLORS_POOL.length];
		}

		timeline = new Timeline(new KeyFrame(Duration.millis(1000 / config.updateFPS), e -> {
			repaint();
		}));
		timeline.setCycleCount(Animation.INDEFINITE);
		timeline.play();

		return group;
	}

	public void repaint() {
		// clear screen
		gc.clearRect(0, 0, canvas.getWidth(), canvas.getHeight());

		// draw funny lines
		for (int i = 0; i < lines.length; ++i) {
			lines[i].updatePosition(1.0 / config.updateFPS);
			if (!lines[i].isAlive()) {
				lines[i].setRandom(r, canvas.getWidth());
			}

			gc.setStroke(lines[i].color);
			gc.setLineWidth(lines[i].width);
			gc.strokeLine(lines[i].x, 10, lines[i].x, canvas.getHeight() - 10);
		}

		gc.setFill(Color.BLACK);
		gc.setStroke(Color.BLACK);
		int x = 20;
		int y = 20;
		writeText("project", x + 3, y, fontItemName);
		// main title
		if (config.projectTitle != null) {
			writeText(config.projectTitle, x, y + 23, fontItemValue);
		}
		// subtitle
		if (config.projectSubtitle != null) {
			writeText(config.projectSubtitle, x + 65, y + 35, fontSubtitle);
		}
		if (config.additionalTitle != null) {
			y += TITLES_GAP;
			writeText(config.additionalTitle, x + 3, y, fontItemName);
			writeText(config.additionalText, x, y + 23, fontItemValue);
		}
		if (config.author != null) {
			y += TITLES_GAP;
			writeText("author", x + 3, y, fontItemName);
			writeText(config.author, x, y + 23, fontItemValue);
		}
		if (config.desc != null) {
			y += TITLES_GAP;
			writeText("desc", x + 3, y, fontItemName);

			gc.setFont(fontDesc);
			y += 23;
			for (int i = 0; i < config.desc.length; ++i) {
				gc.fillText(config.desc[i], x, y);
				y += DESC_GAP;
			}
		}

		if (config.date != null) {
			writeText(config.date, canvas.getWidth() - 60, canvas.getHeight() - 5, fontDate);
		}
	}

	private void writeText(String text, double x, double y, Font font) {
		gc.setFont(font);
		gc.fillText(text, x, y);
	}

	public static void main(String[] args) {
		Application.launch(AboutRunner.class, args);
	}

	public static class AboutRunner extends Application {
		public AboutRunner() {
		}

		@Override
		public void start(Stage primaryStage) throws Exception {
			AboutConfig config = new AboutConfig();
			config.projectTitle = AboutRunner.class.getSimpleName() + " Testing";
			config.projectSubtitle = "version 0.0";
			config.author = "Duffy Duck";
			config.date = "1/2000";
			config.additionalTitle = "developed on";
			config.additionalText = "Windows10, Java8";
			config.desc = new String[] { "Testing app for", "About window.", "", "last line after empty one" };
			config.numberOfLines = 20;
			create(primaryStage, AboutRunner.class.getSimpleName() + " Testing", null, config);
		}
	}

}
