package com.jozki.uutils.javafx.screen;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javafx.animation.Animation;
import javafx.geometry.Point2D;
import javafx.scene.Node;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.Pane;
import javafx.scene.paint.Color;
import javafx.scene.transform.Affine;
import javafx.scene.transform.NonInvertibleTransformException;

public abstract class FxScreen {

	public enum ScreenMode {
		S9_640x350(640, 350), S7_320x200(320, 200);

		public int width;
		public int height;

		private ScreenMode(int width, int height) {
			this.width = width;
			this.height = height;
		}
	}

	private static final int DEFAULT_STRONGNESS = 3;

	private Canvas canvas;
	protected GraphicsContext gc;
	private Pane root;

	private int width;
	private int height;

	private Affine gcTransform;

	private boolean scaled;
	private boolean repaintOnResize;

	private FxScreenEventListener eventListener;
	private Set<KeyCode> exitKeyCodes = new HashSet<>();

	private List<Animation> animations = new LinkedList<>();

	public FxScreen(Pane root) {
		this(root, -1, -1);
	}

	public FxScreen(Pane root, boolean repaintOnResize) {
		this(root, -1, -1, repaintOnResize);
	}

	public FxScreen(Pane root,   ScreenMode mode) {
		this(root,  mode.width, mode.height);
	}

	public FxScreen(Pane root,   int width, int height) {
		this(root,  width, height, true);
	}

	public FxScreen(Pane root,   ScreenMode mode, boolean repaintOnResize) {
		this(root,  mode.width, mode.height, repaintOnResize);
	}

	public FxScreen(Pane root,  int width, int height, boolean repaintOnResize) {
		this.root = root;
		this.width = width;
		this.height = height;
		this.repaintOnResize = repaintOnResize;

		scaled = width > 0 && height > 0;

		canvas = new Canvas();
		gc = canvas.getGraphicsContext2D();

		canvas.widthProperty().addListener(e -> onResize());
		canvas.heightProperty().addListener(e -> onResize());

		canvas.widthProperty().bind(root.widthProperty());
		canvas.heightProperty().bind(root.heightProperty());

		root.getChildren().add(0, canvas);
		
		addEventsHandlers(canvas);
	}

	protected final void exit() {
		exit(null);
	}

	protected final void exit(String command) {
		destroy();
		if (eventListener != null) {
			eventListener.fxScreenExited(this, command);
		}
	}

	public void addExitKeys(KeyCode... keyCodes) {
		for (KeyCode kc : keyCodes) {
			exitKeyCodes.add(kc);
		}
	}

	public boolean isExitKey(KeyCode keyCode) {
		return exitKeyCodes.contains(keyCode);
	}

	public void setEventListener(FxScreenEventListener eventListener) {
		this.eventListener = eventListener;
	}

	protected Canvas getCanvas() {
		return canvas;
	}

	protected void onResize() {
		if (scaled) {
			gc.setTransform(canvas.getWidth() / getWidth(), 0, 0, canvas.getHeight() / getHeight(), 0, 0);
		} else {
			width = (int) canvas.getWidth();
			height = (int) canvas.getHeight();
		}
		if (repaintOnResize) {
			onRepaint();
		}
	}

	protected abstract void onRepaint();

	protected void registerAnimation(Animation animation) {
		if (animation != null) {
			animations.add(animation);
		}
	}

	public void reset() {
		// nop
	}

	public void destroy() {
		for (Animation animation : animations) {
			animation.stop();
		}
		canvas.widthProperty().unbind();
		canvas.heightProperty().unbind();
		root.getChildren().remove(canvas);
	}

	protected void clearScreen() {
		clearScreen(null);
	}

	protected void clearScreen(Color background) {
		if (background != null) {
			gc.setFill(background);
			gc.fillRect(0, 0, getWidth(), getHeight());
		} else {
			gc.clearRect(0, 0, getWidth(), getHeight());
		}
	}

	public int getWidth() {
		return width;
	}

	public int getHeight() {
		return height;
	}

	public void drawPixel(double x, double y, Color color) {
		gc.setStroke(color);
		drawPixel(x, y);
	}

	public void drawPixel(Point2D point, Color color) {
		drawPixel(point.getX(), point.getY(), color);
	}

	public void drawPixel(double x, double y) {
		gc.strokeLine(x, y, x, y);
	}

	public void drawPixel(Point2D point) {
		drawPixel(point.getX(), point.getY());
	}

	public void strokeRect(int x1, int y1, int x2, int y2) {
		gc.strokeLine(x1, y1, x1, y2);
		gc.strokeLine(x1, y2, x2, y2);
		gc.strokeLine(x2, y1, x2, y2);
		gc.strokeLine(x2, y1, x1, y1);
	}

	public void fillRect(int x1, int y1, int x2, int y2) {
		gc.fillRect(x1, y1, x2 - x1, y2 - y1);
	}

	public Point2D transformToCanvas(MouseEvent event) {
		return transformToCanvas(event.getX(), event.getY());
	}

	public Point2D transformToCanvas(Point2D point) {
		return transformToCanvas(point.getX(), point.getY());
	}

	public Point2D transformToCanvas(double x, double y) {
		try {
			return (gcTransform = gc.getTransform(gcTransform)).inverseTransform(x, y);
		} catch (NonInvertibleTransformException e) {
			System.err.println("transformToCanvas: " + e.getMessage());
			return new Point2D(x, y);
		}
	}

	public Point2D transformFromCanvas(Point2D point) {
		return transformFromCanvas(point.getX(), point.getY());
	}

	public Point2D transformFromCanvas(double x, double y) {
		return (gcTransform = gc.getTransform(gcTransform)).transform(x, y);
	}

	private boolean keys[] = new boolean[KeyCode.values().length];

	protected boolean isKey(KeyCode keyCode) {
		return keys[keyCode.ordinal()];
	}

	/**
	 * Overloaded method should return true if key was processed.
	 * @param keyEvent
	 * @return
	 */
	public boolean onKeyPressed(KeyEvent keyEvent) {
		keys[keyEvent.getCode().ordinal()] = true;
		return false;
	}

	public boolean onKeyReleased(KeyEvent keyEvent) {
		keys[keyEvent.getCode().ordinal()] = false;
		if (isExitKey(keyEvent.getCode())) {
			exit();
			return true;
		}
		return false;
	}

	protected void parseAndDraw(String[] commands) {
		for (String command : commands) {
			parseAndDraw(00, 0, command, DEFAULT_STRONGNESS);
		}
	}

	protected void parseAndDraw(int initX, int initY, String command) {
		parseAndDraw(initX, initY, command, DEFAULT_STRONGNESS);
	}

	protected void parseAndDraw(int initX, int initY, String command, int strongness) {
		int x1 = 0;
		int y1 = 0;
		int x2 = 0;
		int y2 = 0;
		String[] cmds = command.split(" ");
		for (String cmd : cmds) {
			boolean drawing = true;
			if (cmd.startsWith("b")) {
				drawing = false;
				cmd = cmd.substring(1);
			}
			if (cmd.startsWith("m")) {
				String[] pos = cmd.substring(1).split(",");
				x2 = Integer.parseInt(pos[0]);
				y2 = Integer.parseInt(pos[1]);
			} else {
				int val = Integer.parseInt(cmd.substring(1));
				if (cmd.startsWith("u")) {
					y2 -= val;
				} else if (cmd.startsWith("d")) {
					y2 += val;
				} else if (cmd.startsWith("r")) {
					x2 += val;
				} else if (cmd.startsWith("l")) {
					x2 -= val;
				} else if (cmd.startsWith("e")) {
					x2 += val;
					y2 -= val;
				} else if (cmd.startsWith("f")) {
					x2 += val;
					y2 += val;
				} else if (cmd.startsWith("g")) {
					x2 -= val;
					y2 += val;
				} else if (cmd.startsWith("h")) {
					x2 -= val;
					y2 -= val;
				}
			}
			if (drawing) {
				for (int i = 0; i < strongness; ++i) {
					gc.strokeLine(initX + x1, initY + y1, initX + x2, initY + y2);
				}
			}
			y1 = y2;
			x1 = x2;
		}
	}

	public void onKeyTyped(KeyEvent keyEvent) {
		// nop
	}
	
	protected void addEventsHandlers(Node node) {
		// nop
	}
}
