package com.jozki.uutils.logging;

import java.io.PrintWriter;
import java.io.StringWriter;

public class Logger {

	public static enum LoggerLevel {
		TRACE("T"), DEBUG("D"), INFO(""), WARN("W"), ERROR("ERR");

		final String txt;

		private LoggerLevel(String txt) {
			this.txt = txt;
		}

		public boolean pass(LoggerLevel level) {
			return ordinal() <= level.ordinal();
		}

		@Override
		public String toString() {
			return txt;
		}
	}

	private String loggerName;
	private LoggerLevel level;

	Logger(String loggerName, LoggerLevel level) {
		this.loggerName = loggerName;
		setLevel(level);
	}

	public void trace(String msg) {
		log(LoggerLevel.TRACE, msg);
	}

	public void trace(Object... obj) {
		log(LoggerLevel.TRACE, obj);
	}

	public void debug(String msg) {
		log(LoggerLevel.DEBUG, msg);
	}

	public void debug(Object... obj) {
		log(LoggerLevel.DEBUG, obj);
	}

	public void info(String msg) {
		log(LoggerLevel.INFO, msg);
	}

	public void info(Object... obj) {
		log(LoggerLevel.INFO, obj);
	}

	public void warn(String msg) {
		log(LoggerLevel.WARN, msg);
	}

	public void warn(Object... obj) {
		log(LoggerLevel.WARN, obj);
	}

	public void error(String msg) {
		log(LoggerLevel.ERROR, msg);
	}

	public void error(Object... objs) {
		log(LoggerLevel.ERROR, objs);
	}

	private void log(LoggerLevel level, Object... objs) {
		if (!this.level.pass(level)) {
			return;
		}

		if (objs == null || objs.length == 0) {
			return;
		}

		if (objs[0] instanceof String) {
			logFormatted(level, (String) objs[0], objs);
		} else {
			logFormatted(level, null, objs);
		}
	}

	private enum State {
		COPY, PARAM_START,
	}

	private void logFormatted(LoggerLevel level, String msg, Object[] params) {
		boolean logException = false;
		if (isLastException(0, params)) {
			logException = true;
		}

		StringBuilder sb = new StringBuilder();
		if (msg != null) {
			State state = State.COPY;
			int paramId = 1;
			for (int i = 0; i < msg.length(); ++i) {
				char c = msg.charAt(i);

				if (c == '{') {
					state = State.PARAM_START;
					continue;
				}
				if (state == State.PARAM_START) {
					if (c == '}') {
						if (logException && paramId + 1 >= params.length) {
							// we are on last param and that is exception -> ignore
							sb.append("{}");
						} else {
							if (paramId < params.length) {
								sb.append(paramToString(params[paramId]));
								++paramId;
							} else {
								// ignore this and all following params
								sb.append("{}");
							}
						}
						state = State.COPY;
						continue;
					} else {
						sb.append('{');
						state = State.COPY;
					}
				}
				sb.append(c);
			}
		} else {
			for (int paramId = 0; paramId < params.length; ++paramId) {
				if (logException && paramId + 1 == params.length) {
					break;
				}
				sb.append(paramToString(params[paramId]));
			}
		}

		if (logException) {
			Throwable t = (Throwable) params[params.length - 1];
			sb.append(System.lineSeparator());
			sb.append(getStackTraceAsString(t));
		}
		LogManager.getInstance().append(loggerName, System.currentTimeMillis(), level, sb.toString());
	}

	private static String getStackTraceAsString(Throwable t) {
		StringWriter sw = new StringWriter();
		t.printStackTrace(new PrintWriter(sw));
		//return "ST:::" + t + "\n" ;
		return sw.toString();
	}

	private boolean isLastException(int firstParamId, Object[] params) {
		return params[params.length - 1] instanceof Throwable && params.length > 1;
	}

	private String paramToString(Object param) {
		if (param == null) {
			return "null";
		}
		return param.toString();
	}

	public String getLoggerName() {
		return loggerName;
	}

	public void setLevel(LoggerLevel level) {
		this.level = level;
	}
}
